import keras
import logging

from .utils import ensure_tf_type, is_numpy
from .tfops_funcs import tf_reshape, tf_rank, tf_concat, tf_shape, tf_cast, tf_image_crop_and_resize, tf_ones, \
    tf_nn_avg_pool, tf_nn_max_pool
import numpy as np
import string
import random
import tensorflow as tf
import keras.backend as K


def convert_maxpool(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert MaxPooling layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.maxpool')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    kernel_shape = params['kernel_shape']
    stride_shape = params['strides']

    pads = params['pads'] if 'pads' in params else [0, 0, 0, 0, 0, 0]
    pad = 'valid'

    if all([shape % 2 == 1 for shape in kernel_shape]) and \
            all([kernel_shape[i] // 2 == pads[i] for i in range(len(kernel_shape))]) and \
            all([shape == 1 for shape in stride_shape]):
        pad = 'same'
        logger.debug('Use `same` padding parameters.')
    else:
        logger.warning('Unable to use `same` padding. Add ZeroPadding2D layer to fix shapes.')
        padding_name = f"{params['cleaned_name']}_maxpool" + '_pad'
        if len(kernel_shape) == 2:
            padding = None

            if len(pads) == 2 and (pads[0] > 0 or pads[1] > 0):
                padding = (pads[0], pads[1])
            elif len(pads) == 4 and (pads[0] > 0 or pads[1] > 0 or pads[2] > 0 or pads[3] > 0):
                padding = ((pads[0], pads[2]), (pads[1], pads[3]))

            if padding is not None:
                padding_layer = keras.layers.ZeroPadding2D(
                    padding=padding,
                    name=padding_name
                )
                layers[padding_name] = input_0 = padding_layer(input_0)
        else:  # 3D padding
            padding_layer = keras.layers.ZeroPadding3D(
                padding=pads[:len(stride_shape)],
                name=padding_name
            )
            layers[padding_name] = input_0 = padding_layer(input_0)
    if len(kernel_shape) == 2:
        pooling = keras.layers.MaxPooling2D(
            pool_size=kernel_shape,
            strides=stride_shape,
            padding=pad,
            name=f"{params['cleaned_name']}_maxpool",
            data_format='channels_first'
        )
    else:
        pooling = keras.layers.MaxPooling3D(
            pool_size=kernel_shape,
            strides=stride_shape,
            padding=pad,
            name=f"{params['cleaned_name']}_maxpool",
            data_format='channels_first'
        )
    ceil_mode = params.get('ceil_mode', False)
    if ceil_mode:
        if pad == 'valid':
            output_shape = ((np.array(input_0.shape[-len(kernel_shape):]) - np.array(kernel_shape)) / np.array(
                stride_shape)) + 1
        else:
            output_shape = np.floor((np.array(input_0.shape[-len(kernel_shape):]) - 1) / np.array(stride_shape)) + 1
        if not np.array([output_shape[i].is_integer() for i in range(len(output_shape))]).all():
            padding = [0 if output_shape[i].is_integer() else stride_shape[i] for i in range(len(kernel_shape))]
            rand_string = ''.join(random.choices(string.ascii_uppercase + string.digits, k=3))
            if len(kernel_shape) == 2:
                layers[node_name + "_pre_" + rand_string] = keras.layers.ZeroPadding2D(
                    ((0, padding[0]), (0, padding[1])), name=f"{params['cleaned_name']}_pre")(input_0)
            else:
                layers[node_name + "_pre_" + rand_string] = keras.layers.ZeroPadding3D(
                    ((0, padding[0]), (0, padding[1]), (0, padding[2])), name=f"{params['cleaned_name']}_pre")(input_0)
            input_0 = layers[node_name + "_pre_" + rand_string]
    layers[node_name] = pooling(input_0)

def convert_global_max_pool(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert GlobalMaxPool layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    tensor_dim = len(input_0.shape)
    if tensor_dim == 3:
        global_pool = keras.layers.GlobalMaxPooling1D(data_format='channels_first',
                                                      name=f"{params['cleaned_name']}_global_max_pool_3")
    elif tensor_dim == 4:
        global_pool = keras.layers.GlobalMaxPooling2D(data_format='channels_first',
                                                      name=f"{params['cleaned_name']}_global_max_pool_4")
    elif tensor_dim == 5:
        global_pool = keras.layers.GlobalMaxPooling3D(data_format='channels_first',
                                                      name=f"{params['cleaned_name']}_global_max_pool_5")
    else:
        raise NotImplementedError("Global max pooling of dims < 3 or dims > 5 is not supported")
    input_0 = global_pool(input_0)
    new_shape = input_0.shape.as_list()
    new_shape = new_shape[1:]
    new_shape.extend([1] * (tensor_dim - 2))
    reshape_layer = keras.layers.Reshape(new_shape, name=f"{params['cleaned_name']}_global_max_pool_reshape")
    input_0 = reshape_layer(input_0)

    layers[node_name] = input_0


def convert_avgpool(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert AvgPooling layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.avgpool')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    kernel_shape = params['kernel_shape']
    stride_shape = params['strides']

    pads = params['pads'] if 'pads' in params else [0, 0, 0, 0, 0, 0]

    if not any(pads):
        pad = 'valid'

    elif all([shape % 2 == 1 for shape in kernel_shape]) and \
            all([kernel_shape[i] // 2 == pads[i] for i in range(len(kernel_shape))]) and \
            all([shape == 1 for shape in stride_shape]):
        pad = 'same'
        logger.debug('Use `same` padding parameters.')
    else:
        pad = 'valid'
        logger.warning('Unable to use `same` padding. Add ZeroPadding2D layer to fix shapes.')
        padding_name = f"{params['cleaned_name']}_avgpool" + '_pad'
        if len(kernel_shape) == 2:
            padding_layer = keras.layers.ZeroPadding2D(
                padding=pads[:len(stride_shape)],
                name=padding_name
            )
        else:  # 3D padding
            padding_layer = keras.layers.ZeroPadding3D(
                padding=pads[:len(stride_shape)],
                name=padding_name
            )
        layers[padding_name] = input_0 = padding_layer(input_0)

    if len(kernel_shape) == 2:
        pooling = keras.layers.AveragePooling2D(
            pool_size=kernel_shape,
            strides=stride_shape,
            padding=pad,
            name=f"{params['cleaned_name']}_avgpool",
            data_format='channels_first'
        )
    elif len(kernel_shape) == 1:
        pooling = keras.layers.AveragePooling1D(
            pool_size=kernel_shape,
            strides=stride_shape,
            padding=pad,
            name=f"{params['cleaned_name']}_avgpool",
            data_format='channels_first'
        )
    else:
        pooling = keras.layers.AveragePooling3D(
            pool_size=kernel_shape,
            strides=stride_shape,
            padding=pad,
            name=f"{params['cleaned_name']}_avgpool",
            data_format='channels_first'
        )
    layers[node_name] = pooling(input_0)


def convert_global_avg_pool(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert GlobalAvgPool layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    tensor_dim = len(input_0.shape)
    if tensor_dim == 3:
        global_pool = keras.layers.GlobalAveragePooling1D(data_format='channels_first',
                                                          name=f"{params['cleaned_name']}_global_avg_pool_3",
                                                          keepdims=True)
    elif tensor_dim == 4:
        global_pool = keras.layers.GlobalAveragePooling2D(data_format='channels_first',
                                                          name=f"{params['cleaned_name']}_global_avg_pool_4",
                                                          keepdims=True)
    elif tensor_dim == 5:
        global_pool = keras.layers.GlobalAveragePooling3D(data_format='channels_first',
                                                          name=f"{params['cleaned_name']}_global_avg_pool_5",
                                                          keepdims=True)
    else:
        raise NotImplementedError("Global average pooling of dims < 3 or dims > 5 is not supported")
    input_0 = global_pool(input_0)
    layers[node_name] = input_0


def convert_topk(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert topk layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    axis = params.get('axis', -1)
    largest = bool(params.get('largest', 1))
    to_sort = bool(params.get('sorted', 1))
    x = layers[node.input[0]]
    k = layers[node.input[1]][0]
    if not is_numpy(k) and not K.is_keras_tensor(k):  # Eager tensor does not serialize well
        k = k.numpy().astype(np.int32)
    if not largest:
        in_tensor = -x
    else:
        in_tensor = x

    def target_layer(composed_input, to_sort=to_sort, axis=axis):
        in_tensor = composed_input[..., :-1]
        k = composed_input[..., -1]
        for i in range(len(k.shape)):
            k = k[0]
        k = tf.cast(k, tf.int32)
        rank = len(in_tensor.shape)

        if axis >= rank - 1 or axis == -1:
            permuted = in_tensor
        else:
            ord_permute = np.arange(rank)
            ord_permute[axis] = rank - 1
            ord_permute[-1] = axis
            permuted = tf.transpose(in_tensor, ord_permute)

        topk_res = tf.math.top_k(permuted, k=k, sorted=to_sort)
        values_pre_permute = topk_res[0]
        indices_pre_permute = topk_res[1]
        topk_concat = tf.stack([values_pre_permute, tf.cast(indices_pre_permute, tf.float32)])
        if axis >= rank - 1 or axis == -1:
            out = topk_concat
        else:
            ord_permute = [0] + (ord_permute + 1).tolist()
            out = tf.transpose(topk_concat, ord_permute)
        return out
    in_shape = tf_shape(in_tensor, tf_name=f"{params['cleaned_name']}_topk_in_shape")
    k_needed_shape_possible_keras_tensor = tf_concat(
            [(in_shape)[:-1],[1]], axis=-1,tf_name=f"{params['cleaned_name']}_topk_k_needed_shape")

    if hasattr(k_needed_shape_possible_keras_tensor, "_inferred_value"): #is keras tensor
        k_needed_shape = k_needed_shape_possible_keras_tensor._inferred_value
    else:
        k_needed_shape = k_needed_shape_possible_keras_tensor
    k_unsqueezed = tf_ones(k_needed_shape, tf_name=f"{params['cleaned_name']}_topk_k_shape")*\
                   tf_cast(k, tf.float32, tf_name=f"{params['cleaned_name']}_topk_k_cast")
    k_reshaped = tf_cast(k_unsqueezed, in_tensor.dtype, tf_name=f"{params['cleaned_name']}_topk_k_reshaped")
    composed_input = tf_concat([in_tensor, k_reshaped], axis=-1,
                               tf_name=f"{params['cleaned_name']}_topk_k_concat")
    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_topk")
    result = lambda_layer(composed_input)
    pos_axis = axis if axis > 0 else in_shape.shape[0]-1
    new_shape = tf_concat([in_shape[:pos_axis], [k], in_shape[pos_axis+1:]], axis=-1,
                          tf_name=f"{params['cleaned_name']}_topk_output_shape")
    values = tf_reshape(result[0], new_shape,
                        tf_name=f"{params['cleaned_name']}_topk_values_reshape")
    indices = tf_reshape(tf_cast(result[1],
                                 tf.int32,
                                 tf_name=f"{params['cleaned_name']}_topk_indices_cast"),
                         new_shape,
                         tf_name=f"{params['cleaned_name']}_topk_indices_reshape")
    if not largest:
        out_tensor = -values
    else:
        out_tensor = values
    layers[params['_outputs'][0]] = out_tensor
    layers[params['_outputs'][1]] = indices


def convert_roi_align(node, params, layers, lambda_func, node_name, keras_name):
    # extract params
    output_height = params.get('output_height', 1)
    output_width = params.get('output_width', 1)
    sampling_ratio = params.get('sampling_ratio', 0)
    spatial_scale = params.get('spatial_scale', 1.0)
    mode = params.get('mode', 'avg')
    if isinstance(mode, bytes):
        mode = mode.decode('utf-8')

    feature_map = layers[node.input[0]]
    rois = layers[node.input[1]]
    batch_indices = layers[node.input[2]]

    adaptive_ratio = False
    if sampling_ratio <= 0:
        sampling_ratio = int((output_height + output_width) / 2)
        adaptive_ratio = True

    rois = rois * spatial_scale
    box_ind = tf_cast(batch_indices, tf.int32, tf_name=f"{params['cleaned_name']}_roi_cast_batch")
    if keras.backend.image_data_format() == 'channels_first':
        fm_shape = tf_shape(feature_map, tf_name=f"{params['cleaned_name']}_roi_hw")[2:]  # H, W
    else:
        raise NotImplementedError("To support channels_last in RoiAlign - need to remove permutes")
    # extract inputs
    x0 = rois[:, 0:1]
    y0 = rois[:, 1:2]
    x1 = rois[:, 2:3]
    y1 = rois[:, 3:4]
    if not adaptive_ratio:
        crop_shape = (
            output_height * sampling_ratio,
            output_width * sampling_ratio,
        )
        spacing_w = (x1 - x0) / tf_cast(crop_shape[1], dtype=tf.float32,
                                        tf_name=f"{params['cleaned_name']}_roi_cast_crop1")
        spacing_h = (y1 - y0) / tf_cast(crop_shape[0], dtype=tf.float32,
                                        tf_name=f"{params['cleaned_name']}_roi_cast_crop0")
        nx0 = (x0 + spacing_w / 2) / tf_cast(fm_shape[1] - 1, dtype=tf.float32,
                                             tf_name=f"{params['cleaned_name']}_roi_cast_fm_1")
        ny0 = (y0 + spacing_h / 2) / tf_cast(fm_shape[0] - 1, dtype=tf.float32,
                                             tf_name=f"{params['cleaned_name']}_roi_cast_fm_0")

        nw = spacing_w * tf_cast(
            crop_shape[1] - 1,
            dtype=tf.float32,
            tf_name=f"{params['cleaned_name']}_roi_cast_crop_2"
        ) / tf_cast(
            fm_shape[1] - 1,
            dtype=tf.float32,
            tf_name=f"{params['cleaned_name']}_roi_cast_crop_3"
        )
        nh = spacing_h * tf_cast(
            crop_shape[0] - 1,
            dtype=tf.float32,
            tf_name=f"{params['cleaned_name']}_roi_cast_crop_3"
        ) / tf_cast(
            fm_shape[0] - 1,
            dtype=tf.float32,
            tf_name=f"{params['cleaned_name']}_roi_cast_crop_4"
        )
    else:
        roi_width = x1 - x0
        roi_height = y1 - y0
        fm_shape_1 = tf_cast(fm_shape[1] - 1, dtype=tf.float32, tf_name=f"{params['cleaned_name']}_roi_cast_fm_1")
        fm_shape_0 = tf_cast(fm_shape[0] - 1, dtype=tf.float32, tf_name=f"{params['cleaned_name']}_roi_cast_fm_0")
        nx0 = x0 / fm_shape_1
        ny0 = y0 / fm_shape_0
        nw = roi_width / fm_shape_1
        nh = roi_height / fm_shape_0

    boxes = tf_concat([ny0, nx0, ny0 + nh, nx0 + nw], axis=1,
                      tf_name=f"{params['cleaned_name']}_roi_concat"
                      )

    permuted_features = keras.layers.Permute([2, 3, 1],
                                             name=f"{params['cleaned_name']}_roi_channels_last"
                                             )(feature_map)  # move to channels last
    cropped_tensor = tf_image_crop_and_resize(
        permuted_features,
        boxes,
        tf.cast(box_ind, dtype=tf.int32),
        crop_size=(
            output_height * sampling_ratio,
            output_width * sampling_ratio,
        ),
        method='bilinear',
        tf_name=f"{params['cleaned_name']}_roi_crop_resize"
    )

    pooled_tensor = None
    if mode.lower() == 'avg':
        pooled_tensor = tf_nn_avg_pool(
            input=cropped_tensor,
            ksize=[1, sampling_ratio, sampling_ratio, 1],
            strides=[1, sampling_ratio, sampling_ratio, 1],
            padding='SAME',
            tf_name=f"{params['cleaned_name']}_roi_avg_pool",
            data_format='NHWC'
        )
    elif mode.lower() == 'max':
        pooled_tensor = tf_nn_max_pool(
            input=cropped_tensor,
            ksize=[1, sampling_ratio, sampling_ratio, 1],
            strides=[1, sampling_ratio, sampling_ratio, 1],
            padding='SAME',
            tf_name=f"{params['cleaned_name']}_roi_max_pool",
            data_format='NHWC'
        )
    else:
        raise ValueError(f"Unknown pooling mode: {mode}")
    layers[node_name] = keras.layers.Permute([3, 1, 2],
                                             name=f"{params['cleaned_name']}_roi_channels_first")(pooled_tensor)



from .upsampling_layers import convert_upsample


def convert_alias_with_name(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Constant layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    layers[node_name] = layers[node.input[0]]


def convert_resize_nearest(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Constant layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    params['scales'] = (params['height_scale'], params['width_scale'])
    params['mode'] = 'nearest'.encode('utf-8')
    convert_upsample(node, params, layers, lambda_func, node_name, keras_name)




from keras.layers import Layer
import tensorflow as tf


class OnnxAbs(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def call(self, inputs, **kwargs):
        x = tf.math.abs(inputs)
        return x



import logging

import keras
import numpy as np
import tensorflow as tf

from .utils import ensure_tf_type
from .tfops_funcs import tf_math_reduce_mean, tf_math_reduce_variance, tf_sqrt, tf_rank, tf_concat, tf_reshape


def convert_batchnorm(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert BatchNorm2d layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.batchnorm2d')
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    if len(node.input) == 5:
        weights = [
            layers[node.input[1]],
            layers[node.input[2]],
            layers[node.input[3]],
            layers[node.input[4]]
        ]
    elif len(node.input) == 3:
        weights = [
            layers[node.input[1]],
            layers[node.input[2]]
        ]
    else:
        raise AttributeError('Unknown arguments for batch norm')

    eps = params['epsilon'] if 'epsilon' in params else 1e-05  # default epsilon
    momentum = params['momentum'] if 'momentum' in params else 0.9  # default momentum

    if isinstance(keras_name, list):
        keras_name = keras_name[0]

    if len(weights) == 2:
        logger.debug('Batch normalization without running averages')
        bn = keras.layers.BatchNormalization(
            axis=1, momentum=momentum, epsilon=eps,
            center=False, scale=False,
            weights=weights,
            name=f"{params['cleaned_name']}_bn"
        )
    else:
        bn = keras.layers.BatchNormalization(
            axis=1, momentum=momentum, epsilon=eps,
            weights=weights,
            name=f"{params['cleaned_name']}_bn"
        )

    layers[node_name] = bn(input_0)


def convert_instancenorm(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert InstanceNorm2d layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    # based on https://github.com/onnx/onnx/blob/main/docs/Operators.md#InstanceNormalization
    logger = logging.getLogger('onnx2keras.instancenorm2d')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    if len(node.input) == 3:
        scale = layers[node.input[1]]
        bias = layers[node.input[2]]
    else:
        raise AttributeError('Unknown arguments for instance norm')

    epsilon = params['epsilon']
    dims_x = len(input_0.shape)
    axis = list(range(2, dims_x))
    var = tf_math_reduce_variance(input_0, axis=axis, keepdims=True, name=None, tf_name=f"{params['cleaned_name']}_var")
    mean = tf_math_reduce_mean(input_0, axis=axis, keepdims=True, name=None, tf_name=f"{params['cleaned_name']}_mean")
    dim_ones = (1,) * (dims_x - 2)
    scale = np.reshape(scale, (-1, *dim_ones))
    bias = np.reshape(bias, (-1, *dim_ones))
    layers[node_name] = (input_0 - mean) * scale / tf_sqrt(var + epsilon, tf_name=f"{params['cleaned_name']}_sqrt")\
                        + bias


def convert_dropout(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Dropout layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.dropout')

    # In ONNX Dropout returns dropout mask as well.
    if isinstance(keras_name, list) and len(keras_name) > 1:
        keras_name = keras_name[0]

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    ratio = params['ratio'] if 'ratio' in params else 0.0
    lambda_layer = keras.layers.Dropout(ratio, name=f"{params['cleaned_name']}_dropout")
    layers[node_name] = lambda_layer(input_0)


def convert_lrn(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert LRN layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.LRN')
    logger.debug('LRN can\'t be tested with PyTorch exporter, so the support is experimental.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    def target_layer(x, depth_radius=params['size'], bias=params['bias'], alpha=params['alpha'], beta=params['beta']):
        import tensorflow as tf
        from keras import backend as K
        data_format = 'NCHW' if K.image_data_format() == 'channels_first' else 'NHWC'

        if data_format == 'NCHW':
            x = tf.transpose(x, [0, 2, 3, 1])

        layer = tf.nn.local_response_normalization(
            x, depth_radius=depth_radius, bias=bias, alpha=alpha, beta=beta
        )

        if data_format == 'NCHW':
            layer = tf.transpose(x, [0, 3, 1, 2])

        return layer

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_lrn")
    layers[node_name] = lambda_layer(input_0)
    lambda_func[keras_name] = target_layer


def convert_layernorm(node, params, layers, lambda_func, node_name, keras_name):
    axis = params.get('axis', -1)
    epsilon = params.get('epsilon', 1e-05)
    stash_type = params.get('stash_type')
    if stash_type is not None:
        raise Exception("LayerNorm stash_type attribute is not implemented")
    input_x = layers[node.input[0]]
    weight = layers[node.input[1]]
    if len(node.input) > 2:
        bias = layers[node.input[2]]
    else:
        bias = None
    center = True if bias is not None else False
    layer_norm = tf.keras.layers.LayerNormalization(
        axis=axis,
        epsilon=epsilon,
        center=center,
        name=f"{params['cleaned_name']}_LayerNorm"
    )
    input_shape = input_x.shape.as_list()
    if input_shape[axis] is None:
        # reshape input such that the axis dim would be non-None (set by weights)
        tf_input_shape = tf.shape(input_x)
        if axis < 0:
            axis = tf_rank(input_x, tf_name=f"{params['cleaned_name']}_LayerNorm_rank")._inferred_value[0] + axis
        tf_new_shape = tf_concat([tf_input_shape[:axis], [weight.shape[0]], tf_input_shape[axis+1:]], axis=-1,
                                 tf_name=f"{params['cleaned_name']}_LayerNorm_new_shape")
        input_x = tf_reshape(input_x, tf_new_shape, tf_name=f"{params['cleaned_name']}_LayerNorm_reshape_none_axis")
    layer_norm.build(input_x.shape)
    if center:
        layer_norm.set_weights([weight, bias])
    else:
        layer_norm.set_weights([weight])
    layers[node_name] = layer_norm(input_x)



import logging

import keras
import tensorflow as tf

from .utils import ensure_tf_type
from .tfops_funcs import tf_multiply, tf_add, tf_clip_by_value, tf_math_erf, tf_math_tanh, tf_math_softplus


def convert_relu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ReLU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    relu = keras.layers.Activation('relu', name=f"{params['cleaned_name']}_relu")
    layers[node_name] = relu(input_0)


def convert_elu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    alpha = params.get('alpha', keras.layers.ELU.__init__.__defaults__[0])
    elu = keras.layers.ELU(alpha=alpha, name=f"{params['cleaned_name']}_elu")
    layers[node_name] = elu(input_0)


def convert_lrelu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert LeakyReLU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    alpha = params.get('alpha', keras.layers.LeakyReLU.__init__.__defaults__[0])
    leakyrelu = keras.layers.LeakyReLU(alpha=alpha, name=f"{params['cleaned_name']}_leakyrelu")
    layers[node_name] = leakyrelu(input_0)


def convert_sigmoid(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Sigmoid activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    sigmoid = keras.layers.Activation('sigmoid', name=f"{params['cleaned_name']}_sigmoid")
    layers[node_name] = sigmoid(input_0)


def convert_tanh(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Tanh activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    tanh = keras.layers.Activation('tanh', name=f"{params['cleaned_name']}_tanh")
    layers[node_name] = tanh(input_0)


def convert_selu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    selu = keras.layers.Activation('selu', name=f"{params['cleaned_name']}_selu")
    layers[node_name] = selu(input_0)


def convert_soft_plus(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf.keras.activations.softplus(input_0)


def convert_soft_sign(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf.keras.activations.softsign(input_0)


def convert_mish(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = input_0 * tf_math_tanh(tf_math_softplus(input_0))


def convert_hard_swish(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    alpha = 1 / 6
    beta = 0.5
    hard_sigmoid = max(0, min(1, alpha * input_0 + beta))
    hard_swish = input_0 * hard_sigmoid
    layers[node_name] = hard_swish


def convert_gelu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert SELU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf.keras.activations.gelu(input_0)


def convert_softmax(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert softmax activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    axis = params.get('axis', keras.layers.Softmax.__init__.__defaults__[0])
    softmax_layer = keras.layers.Softmax(axis=axis, name=f"{params['cleaned_name']}_softmax")
    layers[node_name] = softmax_layer(input_0)
    layers[node_name].set_shape(layers[node_name].shape)


def convert_prelu(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert PReLU activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.prelu')

    if len(node.input) != 2:
        assert AttributeError('Activation layer PReLU should have 2 inputs.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    W = layers[node.input[1]]

    if params['change_ordering']:
        logger.warning('PRelu + change ordering needs to be fixed after TF graph is built.')
        logger.warning('It\'s experimental.')

    shared_axes = [2, 3]

    # for case when W.shape (n,). When activation is used for single dimension vector.
    shared_axes = shared_axes if len(W.shape) > 1 else None

    prelu = keras.layers.PReLU(weights=[W], shared_axes=shared_axes, name=f"{params['cleaned_name']}_prelu")
    layers[node_name] = prelu(input_0)


def convert_hard_sigmoid(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Hard Sigmoid activation layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    alpha = params.get("alpha", 0.2)
    beta = params.get("beta", 0.5)
    # hard sigmoid logic
    x = tf_multiply(input_0, alpha, tf_name=f"{params['cleaned_name']}_multiply")
    x = tf_add(x, beta, tf_name=f"{params['cleaned_name']}_add")
    x = tf_clip_by_value(x, 0., 1., tf_name=f"{params['cleaned_name']}_clip")
    layers[node_name] = x


def convert_erf(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ERF math operation
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for an activation layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf_math_erf(input_0, tf_name=f"{params['cleaned_name']}_erf")



import tensorflow as tf
from keras.layers import Layer


class OnnxSqrt(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def call(self, inputs, **kwargs):
        x = tf.math.sqrt(inputs)
        return x



import logging

import keras
import numpy as np
from keras import backend as K
import tensorflow as tf

from .customonnxlayer.onnxeinsum import OnnxEinsumLayer
from .exceptions import UnsupportedLayer
from .utils import is_numpy, ensure_tf_type, ensure_float
from .tfops_funcs import tf_math_abs, tf_clip_by_value, tf_math_negative, K_mean, tf_math_reduce_prod, \
    tf_math_reduce_min, tf_math_pow, tf_math_sqrt, tf_cast, tf_argmax, tf_expand_dims, tf_math_reciprocal, \
    tf_logical_not, tf_math_sign, tf_math_sin, tf_math_cosh, tf_math_ceil, tf_math_acosh, tf_math_acos, \
    tf_math_asinh, tf_math_asin, tf_math_atanh, tf_math_tan, tf_math_atan, tf_math_sinh, tf_math_less_equal, \
    tf_bitwise_invert, tf_bitwise_bitwise_and, tf_bitwise_bitwise_or, tf_bitwise_bitwise_xor, tf_cos, \
    tf_math_greater, tf_math_greater, tf_math_greater_equal, tf_logical_and, tf_math_logical_xor, tf_math_logical_or, \
    tf_argmin, tf_math_is_inf, tf_math_is_nan, tf_size, tf_not_equal, tf_where, tf_transpose, tf_gather_nd, \
    tf_multiply, tf_image_non_max_suppression, tf_ones_like, tf_stack, tf_concat

# Handle python 2.7 import error
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


def convert_clip(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert clip layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.clip')
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for clip layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    clip_min = params.get('min')
    clip_max = params.get('max')
    if clip_min is None or clip_max is None:
        if len(node.input) == 1:
            raise UnsupportedLayer('Clip without max or min params')
        if len(node.input) > 1 and node.input[1] != '':
            clip_min = float(layers[node.input[1]])
        if len(node.input) == 3 and node.input[2] != '':
            clip_max = float(layers[node.input[2]])

    if clip_min is None and clip_max is None:
        raise UnsupportedLayer('Clip without max or min params')

    if clip_min is None:
        clip_min = tf.float32.min

    if clip_max is None:
        clip_max = tf.float32.max

    if input_0.dtype == tf.int32:
        clip_min = int(clip_min)
        clip_max = int(clip_max)

    layers[node_name] = tf_clip_by_value(input_0, clip_min, clip_max, tf_name=f"{params['cleaned_name']}_clip")


def convert_log(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Log layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for log layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    def target_layer(x):
        import keras.backend as K
        return K.log(x)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_log")
    layers[node_name] = lambda_layer(input_0)
    lambda_func[keras_name] = target_layer


def convert_neg(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Neg layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for log layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    layers[node_name] = tf_math_negative(input_0, tf_name=f"{params['cleaned_name']}_neg")


def convert_exp(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Exp layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for log layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    def target_layer(x):
        import keras.backend as K
        return K.exp(x)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_exp")
    layers[node_name] = lambda_layer(input_0)
    lambda_func[keras_name] = target_layer


def convert_reduce_sum(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reduce sum.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for reduce sum layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    if 'axes' not in params:
        axis = layers[node.input[1]]
    else:
        axis = params['axes']

    keep_dims = True
    if 'keepdims' in params:
        if params['keepdims'] == 0:
            keep_dims = False

    def target_layer(x, axis=axis, keep_dims=keep_dims):
        import keras.backend as K
        return K.sum(x, keepdims=keep_dims, axis=axis)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_reduce_sum")
    layers[node_name] = lambda_layer(input_0)
    layers[node_name].set_shape(layers[node_name].shape)
    lambda_func[keras_name] = target_layer


def convert_reduce_mean(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reduce mean.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for reduce mean layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    param_keepdims = params.get('keepdims', 1)
    keepdims = param_keepdims == 1
    axes = params['axes']
    layers[node_name] = K_mean(input_0, keepdims=keepdims, axis=axes, tf_name=f"{params['cleaned_name']}_mean")


def convert_reduce_max(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reduce max.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for reduce max layer.')
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    def target_layer(x, axis=params.get('axes'), keepdims=params['keepdims']):
        import keras.backend as K
        return K.max(x, keepdims=(keepdims == 1), axis=axis)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_reduce_max")
    layers[node_name] = lambda_layer(input_0)
    layers[node_name].set_shape(layers[node_name].shape)
    lambda_func[keras_name] = target_layer


def convert_reduce_min(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reduce max.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if params.get("axes") is not None:  # opset 13
        axes = params.get("axes")
    elif len(node.input) == 2:
        axes = layers.get(node.input[1])
    else:
        axes = None
    noop_with_empty_axes = bool(params.get("noop_with_empty_axes", False))
    keepdims = params.get("keepdims", True)
    if noop_with_empty_axes and params.get("axes") is None:
        layers[node_name] = layers[node.input[0]]
    else:
        if axes is None:
            layers[node_name] = tf_math_reduce_min(layers[node.input[0]], keepdims=keepdims,
                                               tf_name=f"{params['cleaned_name']}_min")
        else:
            layers[node_name] = tf_math_reduce_min(layers[node.input[0]], axis=axes, keepdims=keepdims,
                                               tf_name=f"{params['cleaned_name']}_min")

def convert_reduce_prod(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert reduce max.
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if params.get("axes") is not None:  # opset 13
        axes = params.get("axes")
    elif len(node.input) == 2:
        axes = layers.get(node.input[1])
    else:
        axes = None  # default is to reduce over all dimensions
    noop_with_empty_axes = bool(params.get("noop_with_empty_axes", False))
    keepdims = bool(params.get("keepdims", True))
    if noop_with_empty_axes and params.get("axes") is None:
        layers[node_name] = layers[node.input[0]]
    else:
        layers[node_name] = tf_math_reduce_prod(layers[node.input[0]],
                                                axis=axes,
                                                keepdims=keepdims,
                                                tf_name=f"{params['cleaned_name']}_reduce")


def convert_pow(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Pow layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 2:
        assert AttributeError('More than 2 inputs for pow layer.')
    layers[node_name] = tf_math_pow(layers[node.input[0]], layers[node.input[1]],
                                    tf_name=f"{params['cleaned_name']}_pow")


def convert_sqrt(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Sqrt layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for sqrt layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf_math_sqrt(input_0, tf_name=f"{params['cleaned_name']}_sqrt")


def convert_split(node, params, layers, lambda_func, node_name, keras_names):
    """
    Convert Split layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for split layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_names[0])
    axis = params.get("axis", 0)
    try:  # onnx opset12
        splits = params["split"]
    except KeyError as e:  # onnx opset 14
        if len(node.input) > 1:
            splits = layers[node.input[1]]
        else:
            if layers[node.input[0]].shape[axis] % 2 != 0:
                raise AttributeError("No splits supplied to the split block but there are uneven number of channels")
            else:
                splits = [layers[node.input[0]].shape[axis] // 2] * 2
    if not isinstance(splits, Iterable):
        # This might not work if `split` is a tensor.
        chunk_size = K.int_size(input_0)[axis] // splits
        splits = (chunk_size,) * splits
    cur = 0
    for i, split in enumerate(splits):
        if len(splits) > 1:
            node_name = params['_outputs'][i]

        def target_layer(x, axis=axis, start_i=cur, end_i=cur + split):
            slices = [slice(None, None)] * len(K.int_shape(x))
            slices[axis] = slice(start_i, end_i)
            return x[tuple(slices)]

        layers[node_name] = target_layer(input_0)
        cur += split


def convert_cast(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Cast layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    logger = logging.getLogger('onnx2keras.cast')

    if len(node.input) != 1:
        assert AttributeError('More than 1 input for cast layer.')

    if is_numpy(layers[node.input[0]]):
        logger.debug('Cast numpy array')

        cast_map = {
            1: np.float32,
            2: np.uint8,
            3: np.int8,
            5: np.int16,
            6: np.int32,
            7: np.int64,
            9: np.bool,
            10: np.float16,
            11: np.double,
        }
        cast_result = layers[node.input[0]]
        result = (layers[node.input[0]] == None)
        if isinstance(result, (bool, np.bool_)) and not result:
            cast_result = cast_map[params['to']](layers[node.input[0]])
        elif not isinstance(result, (bool, np.bool_)) and not np.any(result):
            cast_result = cast_map[params['to']](layers[node.input[0]])
        layers[node_name] = cast_result
    else:
        input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
        check_cast_map = {
            1: tf.float32,
            2: tf.uint8,
            3: tf.int8,
            5: tf.int16,
            6: tf.int32,
            7: tf.int64,
            9: tf.bool,
            10: tf.float16,
            11: tf.double,
        }
        if input_0.dtype == check_cast_map[params['to']] and not isinstance(input_0, (tf.Tensor, np.ndarray)):
            # casting a tensor to the same dtype create placeholder:0 tensor which does not process well in engine
            # trying to ignore the conversion (since its identity) might result in wrong types due to the way
            # keras changes types on serialization and deserialization.
            # So we up-cast to the most informative type then downcast.
            # I'm Sorry.
            if input_0.dtype != tf.double:
                input_0 = tf_cast(input_0, tf.double, tf_name=f"{params['cleaned_name']}_precast")
            else:
                # We can add an If operation to the graph here if needed
                raise NotImplementedError("Does not support tf.double casting into itself")

        def target_layer(x, dtype=params['to'], k_name=f"{params['cleaned_name']}"):
            import tensorflow as tf
            cast_map = {
                1: tf.float32,
                2: tf.uint8,
                3: tf.int8,
                5: tf.int16,
                6: tf.int32,
                7: tf.int64,
                9: tf.bool,
                10: tf.float16,
                11: tf.double,
            }
            return tf_cast(x, cast_map[dtype], tf_name=f'{k_name}_cast')

        layers[node_name] = target_layer(input_0)


def convert_floor(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Floor layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for floor layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)

    def target_layer(x):
        # Floor is absent in keras.backend
        import tensorflow as tf
        return tf.floor(x)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_floor")
    layers[node_name] = lambda_layer(input_0)
    lambda_func[keras_name] = target_layer


def convert_abs(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    res = tf_math_abs(input_0, tf_name=f'{params["cleaned_name"]}_abs')
    layers[node_name] = res


def convert_identity(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert Identity layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for itentity layer.')

    layers[node_name] = layers[node.input[0]]


def convert_argmax(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ArgMax layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for argmax layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    axis = params.get("axis", -1)
    should_keep_dims = params.get("keepdims", True)

    argmax = tf_argmax(input_0, axis=axis, tf_name=f"{params['cleaned_name']}_argmax")
    if should_keep_dims:
        argmax = tf_expand_dims(argmax, axis=axis, tf_name=f"{params['cleaned_name']}_expand")
    layers[node_name] = argmax


def convert_argmin(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ArgMax layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for argmax layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    axis = params.get("axis", -1)
    should_keep_dims = params.get("keepdims", True)

    argmin = tf_argmin(input_0, axis=axis, tf_name=f"{params['cleaned_name']}_argmin")
    if should_keep_dims:
        argmin = tf_expand_dims(argmin, axis=axis, tf_name=f"{params['cleaned_name']}_argmin_unsqueeze")
    layers[node_name] = argmin


def convert_reduce_l2(node, params, layers, lambda_func, node_name, keras_name):
    """
    Convert ReduceL2 layer
    :param node: current operation node
    :param params: operation attributes
    :param layers: available keras layers
    :param lambda_func: function for keras Lambda layer
    :param node_name: internal converter name
    :param keras_name: resulting layer name
    :return: None
    """
    if len(node.input) != 1:
        assert AttributeError('More than 1 input for reduce_l2 layer.')

    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    axis = params.get("axes", [-1])
    keepdims = params.get("keepdims", 0)

    def target_layer(x, axis=axis, keepdims=keepdims):
        import tensorflow as tf
        if isinstance(axis, list) and len(axis) == 1:
            axis = axis[0]
        return tf.norm(x, axis=axis, keepdims=keepdims == 1)

    lambda_layer = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_reduce_l2")
    layers[node_name] = lambda_layer(input_0)
    lambda_func[keras_name] = target_layer


def convert_reciprocal(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf_math_reciprocal(input_0, tf_name=f"{params['cleaned_name']}_reciprocal")


def convert_not(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf_logical_not(input_0, tf_name=f"{params['cleaned_name']}_not")


def convert_less(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]

    if input_1.dtype == input_0.dtype and not isinstance(input_0, (tf.Tensor, np.ndarray)):
        if input_0.dtype != tf.double:
            # To see why this is needed, see inline comments on convert_cast
            input_0 = tf_cast(input_0, dtype=tf.double, tf_name=f"{params['cleaned_name']}_less_cast")
        else:
            raise NotImplementedError("Casting a tensor to itself is not supported")

    def target_layer(y, x=input_0):
        x = tf.cast(x, y.dtype)
        return tf.math.less(x, y)

    lambda_less = keras.layers.Lambda(target_layer, name=f"{params['cleaned_name']}_less")
    less_output = lambda_less(input_1)
    layers[node_name] = less_output


def convert_sign(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_sign(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_sign")


def convert_sin(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_sin(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_sin")


def convert_cosh(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_cosh(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_cosh")


def convert_ceil(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_ceil(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_ceil")


def convert_acosh(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_acosh(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_acosh")


def convert_acos(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_acos(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_acos")


def convert_asinh(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_asinh(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_asinh")


def convert_asin(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_asin(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_asin")


def convert_atanh(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_atanh(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_atanh")


def convert_tan(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_tan(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_tan")


def convert_atan(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_atan(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_atan")


def convert_sinh(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_sinh(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_sinh")


def convert_less_equal(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_less_equal(layers[node.input[0]], layers[node.input[1]],
                                           tf_name=f"{params['cleaned_name']}_less_equal")


def convert_bitwise_not(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_bitwise_invert(tf.cast(layers[node.input[0]], tf.int32),
                                          tf_name=f"{params['cleaned_name']}_bitwise_not")


def convert_bitwise_and(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_bitwise_bitwise_and(layers[node.input[0]], layers[node.input[1]],
                                               tf_name=f"{params['cleaned_name']}_bitwise_and")


def convert_bitwise_or(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_bitwise_bitwise_or(layers[node.input[0]], layers[node.input[1]],
                                              tf_name=f"{params['cleaned_name']}_bitwise_or")


def convert_bitwise_xor(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_bitwise_bitwise_xor(layers[node.input[0]], layers[node.input[1]],
                                               tf_name=f"{params['cleaned_name']}_bitwise_xor")


def convert_cosine(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = ensure_tf_type(layers[node.input[0]], name="%s_const" % keras_name)
    layers[node_name] = tf_cos(input_0, tf_name=f"{params['cleaned_name']}_cos")


def convert_greater(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_greater(layers[node.input[0]], layers[node.input[1]],
                                        tf_name=f"{params['cleaned_name']}_greater")


def convert_greater_equal(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_greater_equal(layers[node.input[0]], layers[node.input[1]],
                                              tf_name=f"{params['cleaned_name']}_greater_equal")


def convert_and(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_logical_and(layers[node.input[0]], layers[node.input[1]],
                                       tf_name=f"{params['cleaned_name']}_and")


def convert_xor(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_logical_xor(layers[node.input[0]], layers[node.input[1]],
                                            tf_name=f"{params['cleaned_name']}_xor")


def convert_or(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_logical_or(layers[node.input[0]], layers[node.input[1]],
                                           tf_name=f"{params['cleaned_name']}_or")


def convert_trilu(node, params, layers, lambda_func, node_name, keras_name):
    x = layers[node.input[0]]
    k = 0
    if len(node.input) > 1:
        k_tensor = layers[node.input[1]]
        try:
            k = int(tf.keras.backend.get_value(k_tensor))
        except:
            k = 0  # fallback if symbolic
    upper = params.get("upper", 1)

    # cannot use tf.experimental.numpy.tril/triu because this is not an eager tensor and we dont know the shape
    def trilu_fn(tensor):
        shape = tf.shape(tensor)
        m, n = shape[-2], shape[-1]
        row_idx = tf.range(m)[:, None]
        col_idx = tf.range(n)[None, :]
        if upper:
            mask = row_idx <= (col_idx - k)
        else:
            mask = row_idx >= (col_idx - k)
        mask = tf.cast(mask, tensor.dtype)
        mask = tf.broadcast_to(mask, shape)
        return tensor * mask

    result = tf.keras.layers.Lambda(trilu_fn, name=keras_name)(x)
    layers[node_name] = result


def convert_cumsum(node, params, layers, lambda_func, node_name, keras_name):
    exclusive = bool(params.get("exclusive", 0))
    reverse = bool(params.get("reverse", 0))
    layers[node_name] = tf.math.cumsum(layers[node.input[0]], layers[node.input[1]],
                                       exclusive=exclusive, reverse=reverse)


def convert_is_inf(node, params, layers, lambda_func, node_name, keras_name):
    if params.get("detect_negative") is not None or params.get("detect_negative") is not None:
        raise AttributeError("Unsupported params detected in isInf conversion: detect_negative/detect_positive")
    layers[node_name] = tf_math_is_inf(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_is_inf")


def convert_is_nan(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_math_is_nan(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_is_nan")


def convert_size(node, params, layers, lambda_func, node_name, keras_name):
    layers[node_name] = tf_size(layers[node.input[0]], tf_name=f"{params['cleaned_name']}_size")


def convert_non_zero(node, params, layers, lambda_func, node_name, keras_name):
    input_tensor = layers[node.input[0]]
    condition = tf_not_equal(
        input_tensor,
        tf.zeros_like(input_tensor),
        tf_name=f"{params['cleaned_name']}_non_zero_unequal"
    )
    nonzero_indices = tf_where(condition, tf_name=f"{params['cleaned_name']}_non_zero_where")
    nonzero_result = tf_transpose(nonzero_indices, tf_name=f"{params['cleaned_name']}_non_zero_transpose")
    nonzero_result = tf_cast(nonzero_result, tf.int32, tf_name=f"{params['cleaned_name']}_non_zero_cast")
    layers[node_name] = nonzero_result
    # tf.experimental.numpy.nonzero(layers[node.input[0]]) was not giving the right results


def convert_gather_nd(node, params, layers, lambda_func, node_name, keras_name):
    input_tensor = layers[node.input[0]]
    indices_tensor = layers[node.input[1]]
    batch_dims = params.get("batch_dims", 0)
    # tesnsorflow implementation of gather_nd, in any case it fails please try also the pseudo_gathernd function here
    # instead. basically it flattens the params and use normal gather to simulate the result of gathernd
    res = tf_gather_nd(input_tensor, indices_tensor, batch_dims=batch_dims,
                       tf_name=f"{params['cleaned_name']}_gather_nd")
    layers[node_name] = res


def pseudo_gathernd(input_tensor, indices_tensor):
    params_shape = input_tensor.shape
    idx_shape = indices_tensor.shape
    idx_dims = idx_shape[-1]
    gather_shape = params_shape[idx_dims:]
    params_flat = tf.reshape(
        input_tensor,
        tf.concat([[-1], gather_shape], axis=0),
    )
    axis_step = tf.math.cumprod(
        params_shape[:idx_dims],
        exclusive=True,
        reverse=True,
    )

    NUMPY_DTYPES_TO_TF_DTYPES = {
        np.dtype('float16'): tf.float16,
        np.dtype('float32'): tf.float32,
        np.dtype('float64'): tf.float64,

        np.dtype('uint8'): tf.uint8,
        np.dtype('uint16'): tf.uint16,
        np.dtype('uint32'): tf.uint32,
        np.dtype('uint64'): tf.uint64,

        np.dtype('int8'): tf.int8,
        np.dtype('int16'): tf.int16,
        np.dtype('int32'): tf.int32,
        np.dtype('int64'): tf.int64,

        np.dtype('bool_'): tf.bool,
    }

    mul = tf.math.multiply(
        indices_tensor,
        tf.cast(
            axis_step,
            dtype=NUMPY_DTYPES_TO_TF_DTYPES[indices_tensor.dtype] \
                if isinstance(indices_tensor.dtype, np.dtype) else indices_tensor.dtype,
        ),
    )
    indices_flat = tf.reduce_sum(
        mul,
        axis=-1,
    )
    result_flat = tf.gather(
        params_flat,
        indices_flat,
    )
    if len(idx_shape) > 0 and len(idx_shape[:-1]) > 0 and idx_shape[:-1][0] is not None:
        pseudo_gathernd_res = tf.reshape(
            result_flat,
            tf.concat([idx_shape[:-1], gather_shape], axis=0),
        )
    else:
        pseudo_gathernd_res = result_flat

    return pseudo_gathernd_res


def convert_nms(node, params, layers, lambda_func, node_name, keras_name):
    scores = layers[node.input[1]]
    boxes = layers[node.input[0]]

    batch_size = boxes.shape[0]

    if batch_size is None:
        raise AttributeError("Onnx2kerras: NMS conversion does not support dynamic batch."
                             "Please change batch to static or remove NMS from model")
    center_point_box = params.get("center_point_box", 0)
    if center_point_box != 0:
        raise AttributeError("Onnx2kerras: We do not support the center_point_box parameter")

    iou_threshold = 0
    score_threshold = float('-inf')
    max_output_size = [2 ** 30]
    if len(node.input) > 2:
        max_output_size = [min(np.squeeze(layers.get(node.input[2], [2 ** 30])), 2 ** 30)]
    if len(node.input) > 3:
        iou_threshold = layers.get(node.input[3], [0])
    if len(node.input) > 4:
        score_threshold = ensure_float(layers.get(node.input[4], float('-inf')))
        if isinstance(score_threshold, np.ndarray):
            score_threshold = score_threshold[0]
    num_classes = scores.shape[1]
    all_results = []
    try:
        iou_threshold = iou_threshold[0]
    except IndexError:  # iou threshold is already a scalar
        pass
    for batch in range(batch_size):
        for c_class in range(num_classes):
            indices = tf_image_non_max_suppression(boxes=boxes[batch],
                                                   scores=scores[batch, c_class],
                                                   max_output_size=tf.cast(max_output_size[0], tf.int32),
                                                   iou_threshold=iou_threshold,
                                                   score_threshold=score_threshold,
                                                   tf_name=f"{params['cleaned_name']}_nms_{batch}_{c_class}")
            ones_indices = tf_ones_like(indices, tf_name=f"{params['cleaned_name']}_nms_ones_{batch}_{c_class}")
            class_tensor = c_class * ones_indices
            batch_tensor = batch * ones_indices
            res = tf_stack([batch_tensor, class_tensor, indices], axis=-1
                           , tf_name=f"{params['cleaned_name']}_nms_stack_{batch}_{c_class}")
            all_results.append(res)
    layers[node_name] = tf_cast(tf_concat(all_results, axis=0, tf_name=f"{params['cleaned_name']}_nms_concat"),
                                dtype=tf.int64,
                                tf_name=f"{params['cleaned_name']}_nms_cast")


def convert_if(node, params, layers, lambda_func, node_name, keras_name):
    if len(layers[node.input[0]].shape) == 0:
        cond = layers[node.input[0]]
    else:
        cond = layers[node.input[0]][0]
    outputs = [layers[node.attribute[i].g.output[0].name] for i in range(2)]
    outputs_dtypes = [output.dtype for output in outputs]
    outputs_numpy_dtypes = [outputs_dtypes[i] if is_numpy(outputs[i]) else outputs_dtypes[i].as_numpy_dtype for i in
                            range(2)]
    if outputs_numpy_dtypes[0] != outputs_numpy_dtypes[1]:
        smallest_idx = np.argmin([np.iinfo(outputs_numpy_dtypes[i]).max for i in range(2)])
        if is_numpy(outputs[smallest_idx]):
            outputs[smallest_idx] = outputs[smallest_idx].astype(outputs_numpy_dtypes[1 - smallest_idx])
        else:
            outputs[smallest_idx] = tf_cast(outputs[smallest_idx], tf.as_dtype(outputs_dtypes[1 - smallest_idx]),
                                            tf_name=f"{params['cleaned_name']}_if_cast")
    in_vec = outputs[0]
    if is_numpy(in_vec):  # if this is a constant it would not be serialized well. connect it to input
        # f_then = lambda x: in_vec
        new_dtype = in_vec.dtype.type

        # The Tf conversion is required to pass args serialization in leap-model-parser
        def get_empty_array(x, dtype=new_dtype, keras_name=keras_name):
            return tf.convert_to_tensor(np.array([]), dtype=new_dtype, name=f'{params["cleaned_name"]}_if_empty_arr')

        if len(in_vec) == 0:  # empty arrays does not serialize well in lambdas.
            then_lambda = get_empty_array
        else:
            then_lambda = lambda x: in_vec
        lambda_layer = tf.keras.layers.Lambda(then_lambda, name=f"{params['cleaned_name']}_if_serizlize_arr_helper")
        if not K.is_keras_tensor(cond):
            raise NotImplementedError(
                "We do not support an if where both the then branch and the in-vector are constants")
        then_output = lambda_layer(cond)  # this assumes
    else:
        then_output = outputs[0]
    layers[node_name] = tf.keras.backend.switch(cond, then_output, outputs[1])


def convert_einsum(node, params, layers, lambda_func, node_name, keras_name):
    input_0 = layers[node.input[0]]
    input_1 = layers[node.input[1]]
    equation = params['equation'].decode('utf-8')

    is_input_0_constant = isinstance(input_0, (tf.Tensor, np.ndarray))
    is_input_1_constant = isinstance(input_1, (tf.Tensor, np.ndarray))
    if is_input_0_constant and is_input_1_constant:
        layers[node_name] = tf.einsum(equation, *[input_0, input_1], name=keras_name)
    elif is_input_0_constant:
        layers[node_name] = OnnxEinsumLayer(equation, input_0, 0)(input_1, name=keras_name)
    elif is_input_1_constant:
        layers[node_name] = OnnxEinsumLayer(equation, input_1, 1)(input_0, name=keras_name)
    else:
        layers[node_name] = OnnxEinsumLayer(equation, None, None)([input_0, input_1], name=keras_name)



import onnx
from onnx2kerastl import onnx_to_keras


def main():
    onnx_file_path = "onnx_models/model.onnx"

    # Load ONNX model
    onnx_model = onnx.load(onnx_file_path)

    # Call the converter (input - is the main model input name, can be different for your model)
    k_model = onnx_to_keras(onnx_model, ['input_1'])


if __name__ == '__main__':
    main()



from typing import List

import keras.backend as K
from keras.layers import Layer


class OnnxReduceMean(Layer):
    def __init__(self, axes: List[int], keepdims: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.axes = axes
        self.keepdims = keepdims

    def call(self, inputs, **kwargs):
        tensor = K.mean(inputs, keepdims=self.keepdims, axis=self.axes)
        return tensor

    def get_config(self):
        config = super().get_config()
        config.update({
            "axes": self.axes,
            "keepdims": self.keepdims,
        })
        return config
